"""
Created on Mon Jul  5 21:09:06 2021

@author: DW
"""
import hashlib
import logging
from pathlib import Path
from warnings import warn

import numpy as np
import pandas as pd


logger = logging.getLogger(__name__)


#%%
class SpectrumReader:
    """
    Reads a clean spectrum from a file Path or str

    with columns "ramanshift" and "intensity".
    Double checks the values
    Sets a hash attribute afterwards
    """

    supported_filetypes = [".txt"]

    spectrum_data_keys = ("ramanshift", "intensity")
    spectrum_keys_expected_values = {
        "ramanshift": {"expected_values": {"min": -95, "max": 3600, "len": 1600}},
        "intensity": {"expected_values": {"min": 0, "max": 1e4, "len": 1600}},
    }
    # expected_ranges = {"ramanshift": (-95, 3600)}
    # expected_length = 1600
    # using slots since there will be many instances of this class

    __slots__ = (
        *("filepath", "max_bytesize", "spectrum", "spectrum_hash", "spectrum_length"),
        *spectrum_keys_expected_values.keys(),
    )

    def __init__(self, filepath: Path, max_bytesize=10 ** 6):

        if not isinstance(filepath, Path):
            if isinstance(filepath, str):
                filepath = Path(filepath)
            else:
                raise TypeError("Argument given is not Path nor str")

        self.filepath = filepath
        self.max_bytesize = max_bytesize

        self.spectrum = pd.DataFrame(data=[], columns=self.spectrum_data_keys)
        if filepath.exists():
            filesize = filepath.stat().st_size
            if filesize < max_bytesize:

                self.spectrum = self.spectrum_parser(self.filepath)
                self.double_check_spectrum_values(
                    self.spectrum, expected_values=self.spectrum_keys_expected_values
                )
        else:
            logger.warning("File does not exist")

        self.spectrum_hash = self.get_hash_text(self.spectrum)
        self.spectrum_length = len(self.spectrum)

        for key in self.spectrum_data_keys:
            setattr(self, key, self.spectrum[key].to_numpy())

    def spectrum_parser(self, filepath: Path):
        """
        Reads data from a file and converts into pd.DataFrame object

        Parameters
        --------
        filepath : Path, str
            file which contains the data of a spectrum

        Returns
        --------
        pd.DataFrame
            Contains the data of the spectrum in a DataFrame with the selected spectrum keys as columns
        """

        spectrum_data = pd.DataFrame()

        suffix = ""
        suffix = filepath.suffix
        if suffix in self.supported_filetypes:
            if suffix == ".txt":

                try:
                    spectrum_data = self.use_np_loadtxt(filepath)

                except Exception as exc:
                    logger.warning(
                        f"Can not complete use_np_loadtxt for:\n{filepath}\n{exc}"
                    )
                    # data = self.read_text(self.filepath)
            elif suffix == ".xlsx":
                # read excel file input
                # IDEA not implemented yet, select columns etc or autodetect
                spectrum_data = pd.read_excel(filepath)

            elif suffix == ".csv":
                # read csv file input
                # IDEA not implemented yet, select columns etc or autodetect
                spectrum_data = pd.read_excel(filepath)

        else:
            logger.warning(f"Filetype {suffix} not supported")

        return spectrum_data

    def use_np_loadtxt(self, filepath, usecols=(0, 1), **kwargs):

        try:
            loaded_array = np.loadtxt(filepath, usecols=usecols, **kwargs)
        except IndexError:
            logger.debug(f"IndexError called np genfromtxt for {filepath}")
            loaded_array = np.genfromtxt(filepath, invalid_raise=False)
        except ValueError:
            logger.debug(f"ValueError called np genfromtxt for {filepath}")
            loaded_array = np.genfromtxt(filepath, invalid_raise=False)
        except Exception as exc:
            logger.warning(f"Can not load data from txt file: {filepath}\n{exc}")
            loaded_array = np.array([])

        spectrum_data = pd.DataFrame()
        if loaded_array.ndim == len(self.spectrum_data_keys):
            try:
                spectrum_data = pd.DataFrame(
                    loaded_array, columns=self.spectrum_data_keys
                )
            except Exception as exc:
                logger.warning(
                    f"Can not create DataFrame from array object: {loaded_array}\n{exc}"
                )
        return spectrum_data

    def double_check_spectrum_values(
        self, spectrum_data: pd.DataFrame, expected_values: dict = {}
    ):
        if all([i in spectrum_data.columns for i in expected_values.keys()]):
            _len = len(spectrum_data)
            for _key, expectations in expected_values.items():

                if expectations:
                    for exp_method, exp_value in expectations.items():
                        _check = False
                        if "len" in exp_method:
                            _spectrum_value = _len
                            _check = np.isclose(_spectrum_value, exp_value, rtol=0.1)

                        elif "min" in exp_method:
                            _spectrum_value = spectrum_data[_key].min()
                            _check = np.isclose(_spectrum_value, exp_value, rtol=0.2)
                            if not _check:
                                _check = exp_value >= _spectrum_value
                        elif "max" in exp_method:
                            _spectrum_value = spectrum_data[_key].max()
                            _check = exp_value <= _spectrum_value
                        else:
                            # not implemented
                            _check = True
                        if not _check:
                            logger.warning(
                                f"The {exp_method} of this spectrum ({_spectrum_value }) does not match the expected values {exp_value}"
                            )
        else:
            logger.error(
                f"The dataframe does not have all the keys from {self.spectrum_data_keys}"
            )

    @staticmethod
    def read_text(filepath, max_bytes=10 ** 6, encoding="utf-8", errors=None):
        """additional read text method for raw text data inspection"""
        _text = "read_text_method"
        filesize = filepath.stat().st_size
        if filesize < max_bytes:
            try:
                _text = filepath.read_text(encoding=encoding, errors=errors)
                # _text.splitlines()
            except Exception as exc:
                # IDEA specify which Exceptions are expected
                _text += "\nread_error"
                logger.warning(f"file read text error => skipped.\n{exc}")
        else:
            _text += "\nfile_too_large"
            logger.warning(f" file too large ({filesize})=> skipped")

        return _text

    def load_data(self, filepath):
        """old method taken out from SpectrumConstructor"""
        # on_lbl="raw"
        # assert self.file.exists(), f'File: "{self.file}" does not exist.'
        # IDEA import file reader class here
        ramanshift, intensity = np.array([]), np.array([])
        i = 0
        while not ramanshift.any() and i < 2000:
            try:
                ramanshift, intensity = np.loadtxt(
                    filepath, usecols=(0, 1), delimiter="\t", unpack=True, skiprows=i
                )
                # Alternative parsing method with pandas.read_csv
                # _rawdf = pd.read_csv(self.file, usecols=(0, 1), delimiter='\t',
                #                     skiprows=i, header =None, names=['ramanshift','intensity'])
                logger.info(
                    f"{self.file} with len rs({len(ramanshift)}) and len int({len(intensity)})"
                )
                # self._read_succes = True
                # self.spectrum_length = len(ramanshift)
                # self.info.update(
                # {"spectrum_length": self.spectrum_length, "skipped_rows": i})
            except ValueError:
                i += 1

    @staticmethod
    def get_hash_text(dataframe, hash_text_encoding="utf-8"):
        text = dataframe.to_string()
        text_hash = hashlib.sha256(text.encode(hash_text_encoding)).hexdigest()
        return text_hash

    def __repr__(self):
        _txt = f"Spectrum({self.filepath.name}, len={self.spectrum_length})"
        return _txt

    def quickplot(self):
        """Plot for quickly checking the spectrum"""
        try:
            self.spectrum.plot(x="ramanshift", y="intensity")
        except TypeError:
            logger.warning(f"No numeric data to plot")



"""
Created on Sun Aug  8 18:20:20 2021

@author: DW
"""
from pathlib import Path
from typing import List, Collection
import logging

logger = logging.getLogger(__name__)

from .filename_parser import PathParser


def make_collection(raman_files: Collection, **kwargs) -> List[PathParser]:

    pp_collection = []
    # _error_parse_filenames
    for file in raman_files:
        try:
            pp_res = PathParser(file, **kwargs)
            pp_collection.append(pp_res)
        except Exception as e:
            logger.warning(
                f"{__name__} make_collection unexpected error for calling PathParser on\n{file}.\n{e}"
            )
            # .append(file)
    pp_collection = sorted(pp_collection)
    return pp_collection
    # self.pp_collection = pp_collection



"""
Created on Thu Jul 22 11:45:34 2021

@author: DW

For this to work, the app needs be installed inside the docker container

"""
from time import sleep

# import raman_fitting
from .api import make_examples

# from ..api import make_examples

if __name__ == "__main__":
    print(
        'Hello Docker World, the raman_fitting "make examples" command from the containter is starting in 3 seconds...'
    )
    sleep(2)
    print("\n...and Go!.....\n")
    make_examples()



# from collections import OrderedDict

import inspect
from functools import partialmethod
from keyword import iskeyword as _iskeyword
from warnings import warn

from lmfit import Parameter, Parameters

# import operator
# from abc import ABC, abstractmethod
from lmfit.models import GaussianModel, LorentzianModel, Model, VoigtModel

# from lmfit import CompositeModel

# print('name: ',__name__,'file: ', __file__, __package__)

if __name__ in ("__main__"):
    from raman_fitting.utils.coordinators import FieldsTracker as FieldsTracker
else:
    from ...utils.coordinators import FieldsTracker

# __all__ = [Base]


#%%
class BasePeakWarning(UserWarning):  # pragma: no cover
    pass


class BasePeak(type):
    """
    Base class for easier definition of typical intensity peaks found in the
    raman spectra.

    The goal of is this metaclass is to be able to more easily write
    peak class definitions (for possible user input). It tries to find three
    fields in the definition, which are required for a LMfit model creation,
    namely: peak_name, peak_type and the param hints.

    peak_name:
        arbitrary name as prefix for the peak
    peak_type:
        defines the lineshape of the peak, the following options are implemented:
        "Lorentzian", "Gaussian", "Voigt"
    params_hints:
        initial values for the parameters of the peak, at least
        a value for the center position of the peak should be given.

    It tries to find these fields in different sources such as: the class definition
    with only class attributes, init attributes or even in the keywords arguments.
    The FieldsTracker class instance (fco) keeps track of the definition in different
    sources and can check when all are ready. If there are multiple sources with definitions
    for the same field than the source with highest priority will be chosen (based on tuple order).
    Each field is a propery which validates the assigments.

    Sort of wrapper for lmfit.model definition.
    Several of these peaks combined are used to make the lmfit CompositeModel
    (composed in the fit_models module), which will be used for the fit.

    --------
    Example usage
    --------

    "Example class definition with attribute definitions"
    class New_peak(metaclass=BasePeak):
        "New peak child class for easier definition"

        param_hints = { 'center': {'value': 2435,'min': 2400, 'max': 2550}}
        peak_type = 'Voigt' #'Voigt'
        peak_name ='R2D2'

    New_peak().peak_model == <lmfit.Model: Model(voigt, prefix='R2D2_')>

    "Example class definition with keyword arguments"

    New_peak = BasePeak('new',
                        peak_name='D1',
                        peak_type= 'Lorentzian',
                        param_hints = { 'center': {'value': 1500}}
    )
    New_peak()


    """

    _fields = ["peak_name", "peak_type", "param_hints"]
    _sources = ("user_input", "kwargs", "cls_dict", "init", "class_name")
    _synonyms = {
        "peak_name": [],
        "peak_type": [],
        "param_hints": ["input_param_settings"],
    }

    PEAK_TYPE_OPTIONS = ["Lorentzian", "Gaussian", "Voigt"]

    # ('value', 'vary', 'min', 'max', 'expr') # optional
    default_settings = {"gamma": {"value": 1, "min": 1e-05, "max": 70, "vary": False}}
    # _intern = list(super.__dict__.keys())+['__module__','_kwargs']
    subclasses = []

    _debug = False

    # def __call__(cls, *args, **kwargs):
    #     print(f'Request received to create an instance of class: {cls}...')
    def __prepare__(name, bases, **kwargs):
        """prepare method only for debugging"""
        if "debug" in kwargs.keys():
            if kwargs.get("debug", False):
                print(f"__prepare ,name {name}, bases{bases}, kwargs {kwargs}")
        # kwargs.update({'debug': False})
        return kwargs

    def __new__(mcls, name, *args, **kwargs):

        # print(
        #             f"Called __new ,_args {_}"
        #         )
        if len(args) == 2:
            # default arg __new__ input
            bases, cls_dict = args
        else:
            bases, cls_dict = (), {}

        if kwargs.get("debug", False):
            mcls._debug = True
        if cls_dict.get("debug", False):
            mcls._debug = True
        if mcls._debug:
            print(
                f"Called __new ({mcls} ),name {name}, bases{bases}, cls_dict {cls_dict.keys()} kwargs {kwargs}"
            )
        # print(f'vars: {vars(cls)}')

        # From namdedtuple source:
        # test input arguments names
        for name_ in [name] + list(kwargs.keys()):
            if type(name_) is not str:
                raise TypeError("Class name and keywords names must be strings")
            if not name_.isidentifier():
                raise ValueError(
                    "Class name and keywords names must be valid "
                    f"identifiers: {name_!r}"
                )
            if _iskeyword(name_):
                raise ValueError(
                    "Class name and keywords names cannot be a " f"keyword: {name_!r}"
                )

        # Init the fco which check and stores the values for each field
        # when fco.status == True then the green light for class initialization is given
        fco = FieldsTracker(fields=mcls._fields, sources=mcls._sources)

        # First thing add input from source kwargs to fco
        fco.multi_store("kwargs", **kwargs)

        # Also add input class name to fco
        fco.multi_store("class_name", **{"peak_name": name})
        # kwargs_ = {k : val for k,val in cls_dict.items() if k in cls_object._fields}
        _cls_dict_field_kwargs = {}
        # Second thing, check class dict for field values,
        # store and replace field values with properties from metaclass
        for field in mcls._fields:
            if field in cls_dict.keys():  # delete field from cls_dict and store in fco
                value = cls_dict[field]
                fco.store("cls_dict", field, value)
                _cls_dict_field_kwargs.update({field: value})
                del cls_dict[field]
            if hasattr(mcls, field):  # store new field property in cls dict
                _obj = getattr(mcls, field)
                if isinstance(
                    _obj, property
                ):  # if mcls has property function with name == field
                    cls_dict[
                        field
                    ] = _obj  # stores property in cls_dict before init so it will be set as a property for cls instance

        # Third: Define the 'new' init for the class, which sets the values from the fco results
        def _init_subclass_replacement(self, *args, **kwargs):
            if self._debug:
                print(f"child __init__ mod called {self}, {kwargs}")
            # super()
            if hasattr(self, "fco"):
                for k, val in self.fco.results.items():
                    setattr(self, k, val["value"])
                    if self._debug:
                        print(f"fco child __init__ setattr called {k}, {val}")

            # if hasattr(self, 'peak_model'):
            setattr(self, "create_peak_model", self.create_peak_model)
            setattr(self, "peak_model", self.create_peak_model())

        # Fourth thing, check class __init__ for setting of field values and delete from cls dict
        if "__init__" in cls_dict.keys():
            cls_init_part_obj = partialmethod(cls_dict["__init__"])

            sig = inspect.signature(cls_dict["__init__"])
            # breakpoint()
            _cls_init_part_obj_funcs = {
                k: val
                for k, val in cls_dict.items()
                if inspect.isfunction(val) and k != "__init__"
            }

            for fname, func in _cls_init_part_obj_funcs.items():
                setattr(cls_init_part_obj, fname, func)
                cls_init_part_obj_dct_keys = set(cls_init_part_obj.__dict__.keys())
                sig = inspect.signature(func)
                try:
                    func(cls_init_part_obj)
                except Exception as e:
                    warn(
                        f"Definition of the __init__ {fname} fails, please redefine init in class, \n{e}",
                        BasePeakWarning,
                    )
            try:
                cls_init_part_obj.func(cls_init_part_obj)
            except AttributeError:
                warn(
                    f"Definition of the __init__ {name} fails, please redefine {fco.missing} in class",
                    BasePeakWarning,
                )
            _cls_init_part_fco_dict = mcls._cleanup_init_dict(
                cls_init_part_obj.__dict__
            )
            fco.multi_store("init", **_cls_init_part_fco_dict)

            cls_dict["_original_init_"] = cls_dict["__init__"]

            del cls_dict["__init__"]
        else:
            pass

        cls_dict["__init__"] = _init_subclass_replacement

        if fco.status:
            pass
            # print(f'Good class definition all values for fields are present: {fco.results}', BasePeakWarning)
        else:
            warn(
                f"Definition for {name} is not complete, please redefine {fco.missing} in class",
                BasePeakWarning,
            )

        if mcls._debug:
            print(
                f"Calling super__new__() ({mcls} ),name {name}, bases{bases}, cls_dict {cls_dict.keys()}"
            )
        cls_object = super().__new__(
            mcls, name, bases, cls_dict
        )  # ,*args, **{**_attrs_found, **kwargs})
        if mcls._debug:
            print(f"Called super__new__() ({mcls} ),cls_object: {cls_object}")
        # setattr(cls_object, "__init__", init_)
        setattr(cls_object, "_fields", mcls._fields)
        setattr(cls_object, "_debug", mcls._debug)
        setattr(cls_object, "fco", fco)
        setattr(cls_object, "PEAK_TYPE_OPTIONS", mcls.PEAK_TYPE_OPTIONS)
        cls_object = mcls._set_other_methods(cls_object)
        return cls_object

    def __init__(self, name, *args, **kwargs):
        # bases, cls_dict removed and kept only *args
        # print(f"__init_ base called ({self} with name '{name}' args {args} and kwargs {kwargs})")
        # subclassess are appended here
        if self not in self.subclasses:
            self.subclasses.append(self)
        # super().__init__(name)
        # print(f"__init_ super called ({self.__init__})")

    @classmethod
    def _cleanup_init_dict(cls, _dict):
        """cleans up the __init__ dictionary from defined class"""
        _dkeys = list(_dict.keys())
        _result = {}
        while _dkeys:
            _dk = _dkeys.pop()
            _kmatch = [
                (i, _dk) for i in cls._synonyms.keys() if i in _dk
            ]  # clean field match
            _synmatch = [
                (k, syn)
                for k, val in cls._synonyms.items()
                for syn in val
                if syn in _dk
            ]  # synonym field match
            # print(_dk, _kmatch,_synmatch)
            if _kmatch:
                _result.update({i[0]: _dict[i[1]] for i in _kmatch})
            elif not _kmatch and _synmatch:
                _result.update({i[0]: _dict[i[1]] for i in _synmatch})
        return _result

    @classmethod
    def _set_other_methods(cls, cls_object):
        """sets other methods found in this baseclass on the defined cls object"""
        _other_methods = [
            i for i in dir(cls) if not i.startswith("_") and not i == "mro"
        ]
        # and not hasattr(cls_object, i)]
        for method in _other_methods:
            # print(f'__new setting method {method}')
            _mcls_obj = getattr(cls, method)
            if method.endswith("__") and not method.startswith("__"):
                method = f"__{method}"

            if not isinstance(_mcls_obj, property):
                # print(f'__new setting {method} on {_mcls_obj} is callable')
                setattr(cls_object, method, _mcls_obj)
        return cls_object

    @property
    def peak_type(self):
        """This property (str) should be assigned and in self.PEAK_TYPE_OPTIONS"""
        return self._peak_type

    @peak_type.setter
    def peak_type(self, value: str):
        """The peak type property should be in PEAK_TYPE_OPTIONS"""
        if not isinstance(value, str):
            raise TypeError(f'The value "{value}" is not a string.')
        value_ = None
        if any([value.upper() == i.upper() for i in self.PEAK_TYPE_OPTIONS]):
            # self.type_to_model_chooser(value)
            value_ = value
        elif any([i.upper() in value.upper() for i in self.PEAK_TYPE_OPTIONS]):
            _opts = [i for i in self.PEAK_TYPE_OPTIONS if i.upper() in value.upper()]
            if len(_opts) == 1:
                value_ = _opts[0]

                warn(
                    f"Peak type misspelling mistake check {value}, forgiven and fixed with {value_}",
                    BasePeakWarning,
                )
            elif len(_opts) > 1:
                raise ValueError(
                    f'Multiple options {_opts} for misspelled value "{value}" in {self.PEAK_TYPE_OPTIONS}.'
                )
            else:
                raise ValueError(
                    f'Multiple options {_opts} for misspelled value "{value}" in {self.PEAK_TYPE_OPTIONS}.'
                )
        else:
            raise ValueError(
                f'The value "{value}" for "peak_type" is not in {self.PEAK_TYPE_OPTIONS}.'
            )
        if value_:
            self._peak_type = value_
            self.fco.store("user_input", "peak_type", value)
            # self.set_peak_model_from_fields()
            self.peak_model = self.create_peak_model()

    @property
    def peak_model(self):
        if not hasattr(self, "_peak_model"):
            self.create_peak_model()
        else:
            return self._peak_model

    @peak_model.setter
    def peak_model(self, value):
        """
        This property is an instance of lmfit.Model,
        constructed from peak_type, peak_name and param_hints setters
        """
        if not isinstance(value, Model):
            self._peak_model = self.create_peak_model()
        else:
            self._peak_model = value

    def create_peak_model(self):
        _peak_model = None
        if self.fco.status:
            if all(hasattr(self, field) for field in self._fields):

                try:
                    create_model_kwargs = dict(
                        peak_name=self.peak_name,
                        peak_type=self.peak_type,
                        param_hints=self.param_hints,
                    )

                    if hasattr(self, "create_model_kwargs"):
                        _orig_kwargs = self.create_model_kwargs

                    _peak_model = LMfitModelConstructorMethods.create_peak_model_from_name_type_param_hints(
                        **create_model_kwargs
                    )
                    self.create_model_kwargs = create_model_kwargs
                    # model = self.make_model_and_set_param_hints(prefix_, peak_type_, param_hints_)
                except Exception as e:
                    print(f"try make models:\n{self}, \n\t {e}")
            else:
                pass
        else:
            pass
            print(f"missing field {self.fco.missing} {self},\n")
        return _peak_model

    @property
    def param_hints(self):
        """This property is dict of dicts and sets the initial values for the parameters"""
        # if hasattr(self, '_peak_model'):
        if hasattr(self, "_param_hints"):
            if isinstance(self._param_hints, Parameters):
                return self._param_hints
            else:
                raise TypeError(
                    f"{self.__class__.__name__} self._param_hints is not instance of Parameters"
                )

    @param_hints.setter
    def param_hints(self, value, **kwargs):
        # print(f'paramhints: val:{value},\nkw:{kwargs}')
        if isinstance(value, Parameters):
            param_hints_ = value
        else:
            dict_ = {}
            if isinstance(value, dict):
                dict_ = {**dict_, **value}
            if kwargs:
                dict_ = {**dict_, **kwargs}
            # print(f'paramhints: {self},\n {dict_}')
            param_hints_ = LMfitModelConstructorMethods.param_hints_constructor(
                param_hints=dict_, default_settings=self.default_settings
            )
        self._param_hints = param_hints_
        self.fco.store("user_input", "param_hints", param_hints_)
        if not isinstance(self, BasePeak):
            self.peak_model = self.create_peak_model()

    @property
    def peak_name(self):
        """This is the name that the peak_model will get as prefix"""
        if self._peak_name:
            if not self._peak_name.endswith("_"):
                self._peak_name = self._peak_name + "_"
        return self._peak_name

    @peak_name.setter
    def peak_name(self, value: str, maxlen=20):
        if len(value) < maxlen:
            prefix_set = value + "_"
            self._peak_name = value
            # print(f"peak_name set:{value}")
            self.fco.store("user_input", "peak_name", value)
            self.peak_model = self.create_peak_model()
            # self.set_peak_model_from_fields()
        else:
            raise ValueError(
                f'The value "{value}" for peak_name is too long({len(value)}) (max. {maxlen}).'
            )

    def repr__(self):
        _repr = f"{self.__class__.__name__}"
        if hasattr(self, "peak_model"):
            _repr += f", {self.peak_model}"
            _param_center = ""
            if self.peak_model:
                _param_center = self.peak_model.param_hints.get("center", {})
            if _param_center:
                _center_txt = ""
                _center_val = _param_center.get("value")
                _center_min = _param_center.get("min", _center_val)
                if _center_min != _center_val:
                    _center_txt += f"{_center_min} < "
                _center_txt += f"{_center_val}"
                _center_max = _param_center.get("max", _center_val)
                if _center_max != _center_val:
                    _center_txt += f" > {_center_max}"
                _repr += f", center : {_center_txt}"
        else:
            _repr += ": no Model set"
        return _repr

    def print_params(self):
        if self.peak_model:
            self.peak_model.print_param_hints()
        else:
            print(f"No model set for: {self}")


#%%

#%%
# LMfitModelConstructorMethods.create_peak_model_from_name_type_param_hints(peak_model= new.peak_model)
# new.peak_name
#%%
class LMfitModelConstructorMethods:

    PARAMETER_ARGS = inspect.signature(Parameter).parameters.keys()

    @classmethod
    def create_peak_model_from_name_type_param_hints(
        cls,
        peak_model: Model = None,
        peak_name: str = None,
        peak_type: str = None,
        param_hints: Parameter = None,
    ):
        if peak_model:
            param_hints_ = peak_model.make_params()
            peak_name_ = peak_model.prefix
            peak_type_ = peak_model.func.__name__
            if peak_name:
                if peak_name != peak_name_:
                    raise Warning("changed name of peak model")
                    peak_model.prefix = peak_name
            else:
                peak_name = peak_name_
            if peak_type:
                if peak_type != peak_type_:
                    raise Warning("changed type of peak model")
                    peak_model = cls.make_model_from_peak_type_and_name(
                        peak_name=peak_name, peak_type=peak_type
                    )
            if param_hints:
                if param_hints != param_hints_:
                    peak_model = cls.set_params_hints_on_model(peak_model, param_hints)
            else:
                peak_model = cls.set_params_hints_on_model(peak_model, param_hints_)
        else:
            if peak_name:
                pass
            else:
                raise Warning(
                    "no peak_name given for create_peak_model, peak_name will be default"
                )
            if peak_type:
                peak_model = cls.make_model_from_peak_type_and_name(
                    peak_name=peak_name, peak_type=peak_type
                )
                if param_hints:
                    peak_model = cls.set_params_hints_on_model(peak_model, param_hints)
        return peak_model

    def make_model_from_peak_type_and_name(peak_type="Lorentzian", peak_name=""):
        """returns the lmfit model instance according to the chosen peak type and sets the prefix from peak_name"""
        model = None
        if peak_type:
            _val_upp = peak_type.upper()
            if "Lorentzian".upper() in _val_upp:
                model = LorentzianModel(prefix=peak_name)
            elif "Gaussian".upper() in _val_upp:
                model = GaussianModel(prefix=peak_name)
            elif "Voigt".upper() in _val_upp:
                model = VoigtModel(prefix=peak_name)
            else:
                raise NotImplementedError(
                    f'This peak type or model "{peak_type}" has not been implemented.'
                )
        return model

    def param_hints_constructor(param_hints: dict = {}, default_settings: dict = {}):
        """
        This method validates and converts the input parameter settings (dict) argument
        into a lmfit Parameters class instance.
        """
        params = Parameters()

        if default_settings:
            try:
                _default_params = [
                    Parameter(k, **val) for k, val in default_settings.items()
                ]
                params.add_many(*_default_params)
            except Exception as e:
                raise ValueError(
                    f"Unable to create a Parameter from default_parameters {default_settings}:\n{e}"
                )
        if not isinstance(param_hints, dict):
            raise TypeError(
                f"input_param_hints should be of type dictionary not {type(param_hints)}"
            )
        else:
            _valid_parlst = []
            for k, val in param_hints.items():
                try:
                    _par = Parameter(k, **val)
                    _valid_parlst.append(_par)
                except Exception as e:
                    raise ValueError(
                        f"Unable to create a Parameter from {k} and {val}:\n{e}"
                    )
            if _valid_parlst:
                try:
                    params.add_many(*_valid_parlst)
                except Exception as e:
                    raise ValueError(
                        f"Unable to add many Parameters from {_valid_parlst}:\n{e}"
                    )
        return params

    @classmethod
    def set_params_hints_on_model(cls, model, param_hints_):
        _error = ""
        if issubclass(model.__class__, Model) and issubclass(
            param_hints_.__class__, Parameters
        ):
            try:
                for pname, par in param_hints_.items():
                    try:
                        _par_hint_dict = {
                            pn: getattr(par, pn, None)
                            for pn in cls.PARAMETER_ARGS
                            if getattr(par, pn, None)
                        }
                        # _par_hint_dict = {k: val for k, val in _par_hint_dict.items() if val}
                        model.set_param_hint(**_par_hint_dict)
                    except Exception as e:
                        _error += f"Error in make_model_hints, check param_hints for {pname} with {par}, {e}"
                # return model
            except Exception as e:
                _error += f"Error in make_model_hints, check param_hints \n{e}"
        else:
            _error += f"TypeError in make_model_hints, check types of model {type(model)} param_hints_{type(param_hints_)}"
        if _error:
            warn("Errors found in setting of param hints: {_error}", BasePeakWarning)
        return model


#%%



import unittest

# import importlib
from importlib import resources
from pathlib import Path

# from raman_fitting.deconvolution_models import first_order_peaks
import raman_fitting
from raman_fitting.datafiles import example_files
from raman_fitting.indexing.filename_parser import PathParser
from raman_fitting.indexing.filename_parser import (
    _extra_sID_name_mapper,
    _extra_overwrite_sID_from_mapper,
)
from raman_fitting.indexing.filename_parser import (
    _extra_sgrpID_name_mapper,
    _extra_overwrite_sgrpID_from_parts,
)

from raman_fitting.indexing.filename_parser_helpers import filestem_to_sid_and_pos

# import pytest


class TestFilenameParser(unittest.TestCase):

    example_parse_expected = {
        "errEMP2_1.txt": ("errEMP2", 1),
        "errTS2_pos1.txt": ("errTS2", 1),
        "Si_spectrum01.txt": ("Si", 1),
        "testDW38C_pos1.txt": ("testDW38C", 1),
        "testDW38C_pos2.txt": ("testDW38C", 2),
        "testDW38C_pos3.txt": ("testDW38C", 3),
        "testDW38C_pos4.txt": ("testDW38C", 4),
        "DW_AB_CD-EF_GE_pos3": ("DW_AB_CD-EF_GE", 3),
        "DW99-pos3": ("DW99", 3),
        "Si": ("Si", 0),
    }

    result_attr = "parse_result"

    def setUp(self):
        _example_path = Path(example_files.__path__[0])
        _example_files_contents = list(Path(_example_path).rglob("*txt"))

        self.datafiles = _example_files_contents
        # list(filter(lambda x: x.endswith('.txt'), _example_files_contents))
        _pathparsers = []
        for fn in self.datafiles:
            _pathparsers.append(PathParser(_example_path.joinpath(fn)))
        self.data_PPs = _pathparsers
        self.empty_PP = PathParser()

        # Make expected results
        # {i.name: (i.parse_result['SampleID'], i.parse_result['SamplePos']) for i in  self.data_PPs}

    def test_PathParser(self):
        self.assertTrue(all(isinstance(i, PathParser) for i in self.data_PPs))
        # Check if instance has results attribute
        self.assertTrue(all(hasattr(i, self.result_attr) for i in self.data_PPs))

    def test_PathParser_empty(self):
        self.assertTrue(hasattr(self.empty_PP, "_flavour"))
        self.assertTrue(hasattr(self.empty_PP, self.result_attr))

    def test_PP_extra_from_map(self):

        for k, val in _extra_sID_name_mapper.items():
            _mapval = _extra_overwrite_sID_from_mapper(k)

            self.assertEqual(_mapval, val)

    def test_PP_extra_from_parts(self):
        self.assertEqual("TEST", _extra_overwrite_sgrpID_from_parts([], "TEST"))

        for k, val in _extra_sgrpID_name_mapper.items():
            emptymap_PP = PathParser(f"{k}/TEST.txt")
            self.assertEqual(
                val,
                _extra_overwrite_sgrpID_from_parts(emptymap_PP.parts, "TEST"),
            )

    def test_PP_parse_filepath_to_sid_and_pos(self):

        for file, _expected in self.example_parse_expected.items():
            self.assertEqual(filestem_to_sid_and_pos(file), _expected)

    # def test_PathParser(self):
    #     _dfpath = Path(__file__).parent.parent.parent / 'src' / 'raman_fitting' / 'datafiles'
    #     _fls = list(_dfpath.rglob('*.txt'))
    #     _res = []
    #     for fn in _fls:
    #         _res.append(PathParser(fn))
    #     sIDs = [i.parse_result['SampleID'] for i in _res]
    #     self.assertEqual(sIDs, self.sIDs_expected)

    # def test_empty(self):
    #     PathParser('')


if __name__ == "__main__":
    unittest.main()
    self = TestFilenameParser()
    self.setUp()



# flake8: noqa
#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri May 14 09:29:16 2021

@author: zmg
"""


def _testing():
    peak1, res1_peak_spec, res2_peak_spec = (
        modname_1,
        fitres_1,
        fitres_2,
    )
    peak1, res1_peak_spec = "1st_6peaks+Si", self._1st["1st_6peaks+Si"]



# flake8: noqa
#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri May 14 09:01:57 2021

@author: zmg
"""

import unittest
from pathlib import Path

import pytest

from raman_fitting.datafiles import example_files
from raman_fitting.processing.spectrum_constructor import (
    SpectrumDataCollection,
    SpectrumDataLoader,
)


class TestSpectrumDataLoader(unittest.TestCase):
    def setUp(self):
        _example_path = Path(example_files.__path__[0])
        _example_files_contents = list(Path(_example_path).rglob("*txt"))

        self.testfile = next(
            filter(lambda x: "testDW38C_pos4" in x.name, _example_files_contents)
        )
        self.errorfile = next(
            filter(lambda x: "wrong" in x.name, _example_files_contents)
        )

    def test_SpectrumDataLoader_empty(self):

        spd = SpectrumDataLoader()
        self.assertEqual(spd.file.name, "empty.txt")

    def test_SpectrumDataLoader_file(self):
        pass
        spd = SpectrumDataLoader(
            self.testfile, run_kwargs=dict(SampleID="testfile", SamplePos=1)
        )
        self.assertEqual(len(spd.register_df), 1600)
        self.assertEqual(len(spd.register_df.columns), 5)

        # self = spcoll


# class SpectrumData():
def _debugging():
    self = TestSpectrumDataLoader()
    pass
    # spectrum_data = SpectrumDataLoader(
    #     file=meannm[-1], run_kwargs=_spectrum_position_info_kwargs, ovv=meangrp
    # )
    # self = spectrum_data
    # self._despike.Z_t
    # self._despike.input_intensity
    # self = self._despike
    # rr = RL.export_collect[0]
    # spec = rr.fitter.spectra_arg._spectra[0]


if __name__ == "__main__":
    unittest.main()



import pandas as pd

from raman_fitting.exporting.plotting import fit_spectrum_plot, raw_data_export

import logging

logger = logging.getLogger(__name__)


class ExporterError(Exception):
    """Error occured during the exporting functions"""


class Exporter:
    """
    The Exporter class handles all the exporting of spectra and models
    into figures and xlsx files.

    """

    def __init__(self, arg, raw_out=True, plot=True, model_names_prefix=["1st", "2nd"]):
        self.raw_out = raw_out
        self.plot = plot
        try:
            self.delegator(arg)
        except ExporterError:
            logger.warning(
                f"{self.__class__.__qualname__} failed export from {type(arg)}"
            )
        except Exception as e:
            logger.error(
                f"{self.__class__.__qualname__} failed export with unexpected error {e}"
            )

    # Exporting and Plotting
    def delegator(self, arg):
        self.fitter = arg
        if "Fitter" in type(arg).__name__:
            self.fitter = arg
            self.split_results()

            if self.raw_out:
                self.raw_export()

            if self.plot:
                self.export_fitting_plotting_models()
        elif isinstance(arg, list):
            # "list" in type([]).__name__:
            # FIXME
            try:
                self.export_from_list(arg)
            except Exception as e:
                logger.error(
                    "f{self.__class__.__qualname__} failed export from list", e
                )
        else:
            logger.warning(
                "f{self.__class__.__qualname__} failed export from unknown arg type {type(arg)}"
            )
            raise ExporterError

    def export_from_list(self, arg):
        fitter_args = [i for i in arg if hasattr(arg, "fitter")]
        if fitter_args:
            FitRes = pd.concat(
                [
                    val.FitParameters
                    for exp in fitter_args
                    for k, val in exp.fitter.FitResults.items()
                ]
            )
            _info = fitter_args[0].fitter.info
            # self.fitter[0].fitter.info
            self.export_fitparams_grp_per_model(FitRes, _info)

    def export_fitparams_grp_per_model(self, FitRes, _info):
        DestGrpDir = _info.get("DestGrpDir")
        grpnm = _info["SampleGroup"]
        for pknm, pkgrp in FitRes.groupby(level=0):
            peak_destpath = DestGrpDir.joinpath(f"{grpnm}_FitParameters_{pknm}")
            pkgrp.dropna(axis=1).to_excel(
                peak_destpath.with_suffix(".xlsx"), index=False
            )

    def raw_export(self):
        raw_data_export(self.fitter.spectra_arg.fitting_spectra)

    def split_results(self):
        pass
        # self._2nd = _2nd
        # _1st = {k:val for k,val in self.fitter.FitResults.items() if k.startswith('1st')}
        # self._1st = _1st

    def export_fitting_plotting_models(self):
        pars1, pars2 = [], []

        _1st = {
            k: val for k, val in self.fitter.FitResults.items() if k.startswith("1st")
        }
        _2nd = {
            k: val for k, val in self.fitter.FitResults.items() if k.startswith("2nd")
        }

        for modname_2, fitres_2 in _2nd.items():

            self.export_xls_from_spec(fitres_2)
            pars2.append(fitres_2.FitParameters)
            for modname_1, fitres_1 in _1st.items():
                self.export_xls_from_spec(fitres_1)
                try:
                    fit_spectrum_plot(
                        modname_1,
                        modname_2,
                        fitres_1,
                        fitres_2,
                        plot_Annotation=True,
                        plot_Residuals=True,
                    )
                except Exception as e:
                    print(
                        f"Error fit_spectrum_plot:{modname_1}, {fitres_1.raw_data_col}.\n {e}"
                    )
                pars1.append(fitres_1.FitParameters)
        return pd.concat(pars1, sort=False), pd.concat(pars2, sort=False)

    def export_xls_from_spec(self, res_peak_spec):
        try:
            # sID = res_peak_spec.extrainfo['SampleID']
            # peak_destpath = res_peak_spec.extrainfo['DestFittingComps.unique()[0].joinpath(f'Model_{res_peak_spec.peak_model}_{sID}')
            # peak_destpath_extra = res_peak_spec.extrainfo.DestFittingComps.unique()[0].joinpath(f'Extra_{res_peak_spec.peak_model}_{sID}')
            res_peak_spec.FitComponents.to_excel(
                res_peak_spec.extrainfo["DestFittingModel"].with_suffix(".xlsx"),
                index=False,
            )
            # res_peak_spec.extrainfo.to_excel(peak_destpath_extra.with_suffix('.xlsx'), index=False)
        except Exception as e:
            print("Error export_xls_from_spec", e)
        # IDEA define fuction for exporting all the indexes _all_index_export
        # index = RamanExport().export_FitParams_Grp(FitParams1, FitParams2, export_info_out, grpnm,sID)
        # all_index.append(index)
        # pars_index = pd.DataFrame(*all_index,columns=list(GrpNames.sGrp_cols[0:2] +('PeakModel','DestPars')))
        # pars_index.to_excel( export_info_out.get('DestGrpDir').joinpath(f'{sGr}_index.xlsx'))



# -*- coding: utf-8 -*-
import textwrap

from setuptools import setup

# all settings for setup are in setup.cfg
setup(
    use_scm_version=True,
    setup_requires=["setuptools_scm", "setuptools_scm_git_archive"],
)
# package_dir={"": "src"},
# )

# "write_to_template": textwrap.dedent(
#             """
#              # coding: utf-8
#              from __future__ import unicode_literals
#              __version__ = {version!r}
#              """,
#         ).lstrip(),



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import argparse
import pathlib
import importlib.metadata


# def _testing():
#     args = parser.parse_args(['-M', 'debug'])
RUN_MODES = ["normal", "testing", "debug", "make_index", "make_examples"]

try:
    _version = importlib.metadata.version("raman_fitting")
except Exception as e:
    _version = "version.not.found"

_version_text = f"\n=== CLI raman_fitting version: {_version} ===\n"
# print(_version_text)


def main():

    """
    The command line interface for raman_fitting
    """

    parser = argparse.ArgumentParser(
        description="Command-line interface for raman_fitting package main."
    )

    parser.add_argument(
        "-M",
        "--run-mode",
        type=str,
        choices=RUN_MODES,
        help="running mode of package, for testing",
        default="normal",
    )

    parser.add_argument(
        "-sIDs",
        "--sampleIDs",
        nargs="+",
        default=[],
        help="Selection of names of SampleIDs from index to run over.",
    )

    parser.add_argument(
        "-sGrps",
        "--samplegroups",
        nargs="+",
        default=[],
        help="Selection of names of sample groups from index to run over.",
    )

    parser.add_argument(
        "--version",
        # action=print(_version_text),
        action="version",
        version="%(prog)s {}".format(_version),
        # const=_version_text,
        help="Prints out the current version of the raman_fitting distribution, via importlib.metadata.version",
    )

    # Execute the parse_args() method
    args = parser.parse_args()

    # import the raman_fitting package
    import raman_fitting as rf

    print(f"CLI args: {args}")
    if args.run_mode == "normal":
        pass
        # _org_index = OrganizeRamanFiles()
        # RL = RamanLoop(_org_index, run_mode ='normal')
    elif args.run_mode.upper() == "DEBUG":
        args.run_mode = args.run_mode.upper()
        # IDEA Add a FAST TRACK for DEBUG
    elif args.run_mode == "testing":
        pass

    _main_run = rf.MainDelegator(**vars(args))

    # return parser



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from collections import namedtuple

# from . import SpectrumCleaner
# from .utils.file_reader import FileReader


def SpectrumWindows():
    windows = {
        "full": (200, 3600),
        "full_1st_2nd": (800, 3500),
        "low": (150, 850),
        "1st_order": (900, 2000),
        "mid": (1850, 2150),
        "2nd_order": (2150, 3380),
        "normalization": (1500, 1675),
    }
    return windows


def SpectrumWindowLimits():
    windows = {
        "full": (20, -20),
        "full_1st_2nd": (20, -20),
        "low": (10, -10),
        "1st_order": (20, -20),
        "mid": (10, -10),
        "2nd_order": (20, -20),
        "normalization": (10, -10),
    }
    return windows


SpecTemplate = namedtuple("Spectrum", "ramanshift intensity label")


class SpectrumTemplate:
    def __init__(self, spec_name="spectrum_info"):
        self.spec_name = spec_name
        self.grp_names()
        self.set_template()

    def grp_names(self):
        sGrp_cols = ("SampleGroup", "SampleID", "FileCreationDate")
        sPos_cols = ("FileStem", "SamplePos", "FilePath")
        spectrum_cols = ("ramanshift", "intensity_raw", "intensity")
        spectrum_info_cols = ("spectrum_length",)
        export_info_cols = (
            "DestGrpDir",
            "DestFittingPlots",
            "DestFittingComps",
            "DestRaw",
        )
        info_cols = (
            sGrp_cols
            + sPos_cols
            + spectrum_cols
            + spectrum_info_cols
            + export_info_cols
        )
        names = {
            "sGrp_cols": sGrp_cols,
            "sPos_cols": sPos_cols,
            "spectrum_cols": spectrum_cols,
            "spectrum_info_cols": spectrum_info_cols,
            "export_info_cols": export_info_cols,
            "all": info_cols,
        }
        Names = namedtuple("GrpNames", names.keys())
        #        output_ary = np.array(output)   # this is your matrix
        #        output_vec = output_ary.ravel() # this is your 1d-array
        self.grp_names = Names(**names)

    def set_template(self):
        # name = 'spectrum_info'
        self.template = namedtuple(self.spec_name, self.grp_names.all)



import datetime as dt
import logging
from collections import OrderedDict, namedtuple

import pandas as pd

from ..processing.spectrum_constructor import SpectrumDataCollection, SpectrumDataLoader
from .base_model import InitializeModels

logger = logging.getLogger(__name__)


class Fitter:
    """
    Fitter class for executing the fitting functions and optimizations

    IDEA: implement sensing of spectrum for Si samples
    """

    fit_windows = ["1st_order", "2nd_order"]

    def __init__(self, spectra_arg, RamanModels=InitializeModels(), start_fit=True):
        self._qcnm = self.__class__.__qualname__
        logger.debug(f"{self._qcnm} is called with spectrum\n\t{spectra_arg}\n")
        self.start_fit = start_fit
        self.models = RamanModels

        self.spectra_arg = spectra_arg
        self.spectra = spectra_arg
        self.fit_delegator()

    @property
    def spectra(self):
        return self._spectra

    @spectra.setter
    def spectra(self, value):
        """Checks if value is dict or else takes a dict from class instance value"""

        _errtxt = f"This assignment {value} does not contain valid spectra"
        if isinstance(value, dict):
            _data = value
        elif isinstance(value, SpectrumDataCollection):
            _data = value.mean_data
            _fit_lbl = "mean"
        elif isinstance(value, SpectrumDataLoader):
            _data = value.clean_df
            _fit_lbl = "int"
        elif isinstance(value, pd.DataFrame):
            raise AttributeError
            # IDEA implement self.sense_windowname(value)
        else:
            raise ValueError(_errtxt)

        _specs = {
            k: val
            for k, val in _data.items()
            if k in self.fit_windows and type(val) == pd.DataFrame
        }
        # assert bool(_specs), _errtxt
        if not _specs:
            self.start_fit = False

        self._spectra = _specs
        self.FitResults = {}
        info = {}
        if hasattr(value, "info"):
            info = {**info, **value.info}
        self.info = info

    def fit_delegator(self):

        if self.start_fit:
            logger.info(
                f"\n{self._qcnm} is starting to fit the models on spectrum:\n\t{self.info.get('SampleID','no name')}"
            )

            self.fit_models(self.models.second_order)  # second order should go first
            logger.info(
                f"\t - second order finished, {len(self.models.second_order)} model"
            )
            # rum:\t{self.info.get('SampleID','no name')}\n")
            self.fit_models(self.models.first_order)
            logger.info(
                f"\t - first order finished, {len(self.models.first_order)} models"
            )

    def fit_models(self, model_selection):

        _fittings = {}
        logger.debug(f"{self._qcnm} fit_models starting.")
        for modname, model in model_selection.items():
            modname, model
            _windowname = [i for i in self.fit_windows if modname[0:3] in i][0]
            _data = self.spectra.get(_windowname)
            _int_lbl = self.get_int_label(_data)
            try:
                out = self.run_fit(
                    model.lmfit_model,
                    _data,
                    _int_lbl=_int_lbl,
                    _modelname=modname,
                    _info=self.info,
                )
                prep = PrepareParams(out, extra_fit_results=self.FitResults)
                _fittings.update({modname: prep.FitResult})
            except Exception as e:
                logger.warning(
                    f"{self._qcnm} fit_model failed for {modname}: {model}, because:\n {e}"
                )
            
        self.FitResults.update(**_fittings)

    def run_fit(self, model, _data, method="leastsq", **kws):
        # ideas: improve fitting loop so that starting parameters from modelX and modelX+Si are shared, faster...
        _fit_res, _param_res = {}, {}
        init_params = model.make_params()
        x, y = _data.ramanshift, _data[kws.get("_int_lbl")]
        out = model.fit(y, init_params, x=x, method=method)  # 'leastsq'
        for k, val in kws.items():
            if not hasattr(out, k):
                _attrkey = k
            elif not hasattr(out, f"_{k}"):
                _attrkey = f"_{k}"
            else:
                _attrkey = None
            if _attrkey:
                setattr(out, _attrkey, val)
        return out

    def get_int_label(self, value):
        _lbl = ""
        if isinstance(value, pd.DataFrame):
            cols = [i for i in value.columns if "ramanshift" not in i]
            if len(cols) == 0:
                _lbl = ""
            if len(cols) == 1:
                _lbl = cols[0]
            elif len(cols) > 1:
                if any("mean" in i for i in cols):
                    _lbl = [i for i in cols if "mean" in i][0]
                elif any("int" in i for i in cols):
                    _lbl = [i for i in cols if "int" in i][0]

        return _lbl


class PrepareParams:

    fit_attr_export_lst = (
        "chisqr",
        "redchi",
        "bic",
        "aic",
        "method",
        "message",
        "success",
        "nfev",
    )
    fit_result_template = namedtuple(
        "FitResult",
        [
            "FitComponents",
            "FitParameters",
            "FitReport",
            "extrainfo",
            "model_name",
            "raw_data_col",
        ],
    )
    ratio_params = [("I", "_height"), ("A", "_amplitude")]
    _standard_2nd_order = "2nd_4peaks"

    def __init__(self, model_result, extra_fit_results={}):
        self._qcnm = self.__class__.__qualname__
        logger.debug(f"{self._qcnm} is called with model_result\n\t{model_result}\n")
        self.extra_fit_results = extra_fit_results
        self.model_result = model_result

    @property
    def model_result(self):
        return self._model_result

    @model_result.setter
    def model_result(self, value):
        """
        Takes the ModelResult class instance from lmfit.
        Optional extra functionality with a list of instances.
        """
        self.result = {}

        if "ModelResult" in type(value).__name__:
            self.result.update(value.params.valuesdict())
            self.comps = value.model.components
        elif ("list" or "tuple") in type(value).__name__:
            assert all("ModelResult" in type(i).__name__ for i in value)
            [self.result.update(mod.params.valuesdict()) for mod in value]
            self.comps = [i for mod in value for i in mod.model.components]

        self.peaks = set(
            [i.prefix for i in self.comps]
        )  # peaks is prefix from components

        _mod_lbl = "Model"
        if hasattr(value, "_modelname"):
            _mod_lbl = f'Model_{getattr(value,"_modelname")}'
        self.model_name_lbl = _mod_lbl
        
        self.raw_data_lbl = value._int_lbl

        self._model_result = value
        
        self.make_result()

    def make_result(self):
        self.prep_params()
        self.prep_components()
        self.FitReport = self.model_result.fit_report(show_correl=False)

        self.extra_info = {}
        self.prep_extra_info()
        self.FitResult = self.fit_result_template(
            self.FitComponents,
            self.FitParameters,
            self.FitReport,
            self.extra_info,
            self.model_name_lbl,
            self.raw_data_lbl,
        )

    def prep_extra_info(self):
        self.extra_info = {}
        _destfitcomps = self.model_result._info["DestFittingComps"]
        _model_destdir = _destfitcomps.joinpath(
            f'{self.model_name_lbl}_{self.model_result._info["SampleID"]}'
        )
        self.extra_info = {
            **self.model_result._info,
            **{"DestFittingModel": _model_destdir},
        }

    def prep_params(self):
        fit_attrs = OrderedDict(
            zip(
                [f"lmfit_{i}" for i in self.fit_attr_export_lst],
                [getattr(self.model_result, i) for i in self.fit_attr_export_lst],
            )
        )
        self.result.update(fit_attrs)
        try:
            self.add_ratio_params()
        except Exception as e:
            logger.error(f"{self._qcnm} extra prep params failed\n\t{e}\n")

        self.result.update(
            {"_run_date_YmdH": dt.datetime.now().strftime(format="%Y-%m-%d %H:00")}
        )
        self.FitParameters = pd.DataFrame(self.result, index=[self.model_name_lbl])

    def add_ratio_params(self):
        # peaks = [i.prefix for i in self.out.model.components]
        RatioParams = {}
        for a, t in self.ratio_params:
            if {"G_", "D_"}.issubset(self.peaks):
                RatioParams.update(
                    {f"{a}D/{a}G": self.result["D" + t] / self.result["G" + t]}
                )
                RatioParams.update(
                    {f"La_{a}G": 4.4 * RatioParams.get(f"{a}D/{a}G") ** -1}
                )
                #                , 'ID/IG' : fit_params_od['D_height']/fit_params_od['G_height']}
                if "D2_" in self.peaks:
                    RatioParams.update(
                        {
                            f"{a}D/({a}G+{a}D2)": self.result["D" + t]
                            / (self.result["G" + t] + self.result["D2" + t])
                        }
                    )
                    RatioParams.update(
                        {
                            f"La_{a}G+D2": 4.4
                            * RatioParams.get(f"{a}D/({a}G+{a}D2)") ** -1
                        }
                    )
                    #               : fit_params_od['D'+t]/(fit_params_od['G'+t]+fit_params_od['D2'+t])})
                    if "D3_" in self.peaks:
                        RatioParams.update(
                            {
                                f"{a}D3/({a}G+{a}D2": self.result["D3" + t]
                                / (self.result["G" + t] + self.result["D2" + t])
                            }
                        )
            if "D3_" in self.peaks:
                RatioParams.update(
                    {f"{a}D3/{a}G": self.result["D3" + t] / self.result["G" + t]}
                )
            if "D4_" in self.peaks:
                RatioParams.update(
                    {f"{a}D4/{a}G": self.result["D4" + t] / self.result["G" + t]}
                )

            if {"D1D1_", "GD1_"}.issubset(self.peaks):
                RatioParams.update(
                    {
                        f"{a}D1D1/{a}GD1": self.result["D1D1" + t]
                        / self.result["GD1" + t]
                    }
                )
            if self.extra_fit_results:
                RatioParams.update(self.add_ratio_combined_params(a, t))
        self.ratio_params = RatioParams
        self.result.update(RatioParams)

    def add_ratio_combined_params(self, a, t):
        _2nd = self._standard_2nd_order
        if (
            self.model_result._modelname.startswith("1st")
            and _2nd in self.extra_fit_results.keys()
        ):
            _D1D1 = self.extra_fit_results[_2nd].FitParameters.loc[
                f"Model_{_2nd}", "D1D1" + t
            ]
            self.result.update({"D1D1" + t: _D1D1})
            return {f"Leq_{a}": 8.8 * _D1D1 / self.result["D" + t]}
        else:
            return {}

    def prep_components(self):
        # FittingParams = pd.DataFrame(fit_params_od,index=[peak_model])
        _fit_comps_data = OrderedDict({"RamanShift": self.model_result.userkws["x"]})
        _fit_comps_data.update(self.model_result.eval_components())
        
        # IDEA take out
        # print('===/n',self.model_result, '/n')
        # print('===/n',self.model_result.__dict__.keys(), '/n')
        
        
        _fit_comps_data.update(
            {
                self.model_name_lbl: self.model_result.best_fit,
                "residuals": self.model_result.residual,
                self.model_result._int_lbl: self.model_result.data,
            }
        )
        FittingComps = pd.DataFrame(_fit_comps_data)
        self.FitComponents = FittingComps


def NormalizeFit(norm_cleaner, plotprint=False):  # pragma: no cover
    # IDEA: optional add normalization seperately to Fitter
    x, y = norm_cleaner.spec.ramanshift, norm_cleaner.blcorr_desp_intensity
    Model = InitializeModels("2peaks normalization Lorentzian")
    params = Model.make_params()
    pre_fit = Model.fit(y, params, x=x)  # 'leastsq'
    IG, ID = pre_fit.params["G_height"].value, pre_fit.params["D_height"].value
    output = {
        "factor": 1 / IG,
        "ID/IG": ID / IG,
        "ID": ID,
        "IG": IG,
        "G_center": pre_fit.params["G_center"].value,
        "D_center": pre_fit.params["D_center"].value,
        "Model": Model,
    }
    #    pre_fit = Model.fit(y,params ,x=x,method='differential-evolution') # 'leastsq'
    if plotprint:
        pre_fit.plot()
        print(pre_fit.fit_report())
    return output
